#ifndef __JSON_H__
#define __JSON_H__

#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>

typedef struct JsonString {
  uint8_t *buf;
  uintptr_t len;
} JsonString;

typedef struct Pair {
  struct JsonString key;
  struct Json *val;
} Pair;

/**
 * Json struct for C
 */
typedef enum Json_Tag {
  Array,
  Object,
  String,
  Number,
  True,
  False,
  Null,
  Error,
} Json_Tag;

typedef struct Array_Body {
  struct Json *elems;
  uintptr_t len;
} Array_Body;

typedef struct Object_Body {
  struct Pair *elems;
  uintptr_t len;
} Object_Body;

typedef struct Json {
  Json_Tag tag;
  union {
    Array_Body array;
    Object_Body object;
    struct {
      struct JsonString string;
    };
    struct {
      double number;
    };
  };
} Json;

/**
 * Configures the JSON parser
 */
typedef struct JsonConfig {
  /**
   * Max depth for nested objects
   */
  uint32_t max_depth;
  /**
   * Recover from errors.
   * For example, trailing commas on objects
   * are not allowed, but this flag makes
   * the parser skip them.
   */
  bool recover_from_errors;
} JsonConfig;

/**
 * Deserializes the given string into a Json struct.
 * If any error is encountered while parsing, the
 * type of the Json struct is `Json::Error`.
 *
 * The caller of this function must free the returned
 * struct by calling [`json_free`] afterwards.
 *
 * # Safety
 * The pointer must be a valid NULL terminated C string
 */
struct Json json_deserialize(const char *ptr);

/**
 * Deserializes the given string into a Json struct.
 * Using the given `JsonConfig` struct
 *
 * If any error is encountered while parsing, the
 * type of the Json struct is `Json::Error`.
 *
 * The caller of this function must free the returned
 * struct by calling [`json_free`] afterwards.
 *
 * # Safety
 * The pointer must be a valid NULL terminated C string
 */
struct Json json_deserialize_with_config(const char *ptr, struct JsonConfig conf);

/**
 * Frees the given Json structure.
 */
void json_free(struct Json json);

#endif /* __JSON_H__ */
